"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.src.applications.xception import Xception as Xception
from keras.src.applications.xception import (
    decode_predictions as decode_predictions,
)
from keras.src.applications.xception import preprocess_input as preprocess_input
